"""Tests for the settings module"""

import os
from typing import Type

import pytest

from dkist_service_configuration import ConfigurationBase
from dkist_service_configuration import MeshServiceConfigurationBase
from dkist_service_configuration.settings import DEFAULT_MESH_SERVICE
from dkist_service_configuration.settings import InstrumentedMeshServiceConfigurationBase


@pytest.fixture()
def environment_vars() -> dict:
    os.environ["foo"] = "1"
    os.environ["bar"] = "2"
    os.environ["LOGURU_LEVEL"] = "WARNING"
    os.environ[
        "MESH_CONFIG"
    ] = """{
    "upstream_service_name": {
        "mesh_address": "localhost",
        "mesh_port": 6742
    },
    "otlp_traces_endpoint": {"mesh_address": "127.0.0.1", "mesh_port": 4317},
    "otlp_metrics_endpoint": {"mesh_address": "127.0.0.1", "mesh_port": 4317}
}"""
    return dict(os.environ)


@pytest.fixture()
def base_config() -> Type[ConfigurationBase]:
    class BaseConfig(ConfigurationBase):
        foo: str

    return BaseConfig


def test_base_config(environment_vars, base_config):
    config = base_config()
    assert config.log_level == "WARNING"
    config.log_configurations()
    md_table = config.model_fields_to_markdown_table()
    print(md_table)
    rst_table = config.model_fields_to_restructured_text_table()
    print(rst_table)


@pytest.fixture()
def mesh_config() -> Type[MeshServiceConfigurationBase]:
    class MeshConfig(MeshServiceConfigurationBase):
        bar: str = "foo"
        password: str = "bar"

    return MeshConfig


def test_mesh_config(environment_vars, mesh_config):
    config = mesh_config()
    assert config.log_level == "WARNING"
    assert config.service_mesh_detail("upstream_service_name").host == "localhost"
    assert config.service_mesh_detail("upstream_service_name").port == 6742
    config.log_configurations()


def test_default_mesh_config(environment_vars, mesh_config):
    config = mesh_config()
    assert config.service_mesh_detail("not_a_configured_service").host == DEFAULT_MESH_SERVICE.host
    assert config.service_mesh_detail("not_a_configured_service").port == DEFAULT_MESH_SERVICE.port


def test_instrumented_mesh_service_configuration_tracer(instrumented_mesh_config):
    """
    Given an instrumented mesh service configuration
    When the tracer is initialized
    Then the tracer should be set up correctly
    """
    tracer = instrumented_mesh_config.tracer
    with tracer.start_as_current_span(
        name="test_instrumented_mesh_service_configuration_tracer"
    ) as span:
        pass


def test_instrumented_mesh_service_configuration_meter(instrumented_mesh_config):
    """
    Given an instrumented mesh service configuration
    When the meter is initialized
    Then the meter should be set up correctly
    """
    meter = instrumented_mesh_config.meter
    counter = meter.create_up_down_counter(
        name="test_instrumented_mesh_service_configuration_meter",
        unit="1",
        description="The number of times this counter is incremented",
    )
    counter.add(1)


def test_auto_instrument(instrumented_mesh_config):
    """
    Given an instrumented mesh service configuration
    When the auto_instrument method is called
    Then the tracer and meter should be set up correctly
    """
    instrumented_mesh_config.auto_instrument()
    assert instrumented_mesh_config.tracer is not None
    assert instrumented_mesh_config.meter is not None



"""Unit tests for the message_processor module."""

from unittest.mock import MagicMock

import pika
import pytest
from talus import ConsumeMessageBase

from dkist_service_configuration.message_processor import InstrumentedMessageProcessorBase


@pytest.fixture()
def channel() -> pika.adapters.blocking_connection.BlockingChannel:
    return MagicMock(spec=pika.adapters.blocking_connection.BlockingChannel)


@pytest.fixture()
def properties() -> pika.spec.BasicProperties:
    return pika.spec.BasicProperties()


@pytest.fixture()
def method() -> pika.spec.Basic.Deliver:
    return pika.spec.Basic.Deliver(routing_key="test.m", delivery_tag="test")


@pytest.fixture()
def valid_body() -> dict:
    data = {
        "bucket": "inbox",
        "objectName": "bar",
    }
    return data


@pytest.fixture()
def message_processor_cls() -> InstrumentedMessageProcessorBase:
    class MessageProcessor(InstrumentedMessageProcessorBase):
        def process_message(self, message: ConsumeMessageBase):
            self._process_message_called = True

    return MessageProcessor


def test_message_processor_success(
    channel,
    method,
    properties,
    valid_body,
    message_processor_cls,
):
    """
    :given: A message processor that does not raise an exception.
    :when: The message processor is called.
    :then: The message processor should process and acknowledge the message.
    """
    # given
    message_processor = message_processor_cls()
    # when
    message_processor(channel=channel, method=method, properties=properties, body=valid_body)
    # then
    assert message_processor._process_message_called
    channel.basic_ack.assert_called_once_with(delivery_tag=method.delivery_tag)


@pytest.fixture()
def invalid_body() -> str:
    return "not good, json {}"


def test_message_processor_handled_error(
    channel,
    method,
    properties,
    invalid_body,
    message_processor_cls,
):
    """
    :given: A message processor that raises an exception.
    :when: The message processor is called.
    :then: The message processor should process and reject the message.
    """
    # given
    message_processor = message_processor_cls()
    # when
    message_processor(channel=channel, method=method, properties=properties, body=invalid_body)
    # then
    channel.basic_nack.assert_called_once_with(delivery_tag=method.delivery_tag, requeue=False)


@pytest.fixture()
def erroring_message_processor_cls() -> InstrumentedMessageProcessorBase:
    class MessageProcessor(InstrumentedMessageProcessorBase):
        def process_message(self, message: ConsumeMessageBase):
            self._process_message_called = True
            raise RuntimeError("Error")

    return MessageProcessor


def test_message_processor_unhandled_error(
    channel,
    method,
    properties,
    valid_body,
    erroring_message_processor_cls,
):
    """
    :given: A message processor that does raise an exception.
    :when: The message processor is called.
    :then: The message processor should reraise the error.
    """
    # given
    message_processor = erroring_message_processor_cls()
    # when/then
    with pytest.raises(RuntimeError):
        message_processor(channel=channel, method=method, properties=properties, body=valid_body)



"""Tests for the logging module"""

import logging

import pytest

from dkist_service_configuration.logging import format_log_level_from_record
from dkist_service_configuration.logging import logger

std_logger = logging.getLogger(__name__)


def test_log_levels():
    logger.trace("trace")
    logger.debug("debug")
    logger.info("info")
    logger.warning("warning")
    logger.error("error")
    logger.critical("critical")


def test_log_levels_std_logger():
    std_logger.debug("debug")
    std_logger.info("info")
    std_logger.warning("warning")
    std_logger.error("error")


@pytest.mark.parametrize(
    "level, expected",
    [
        ("TRACE", 5),
        ("DEBUG", 10),
        ("INFO", 20),
        ("SUCCESS", 25),
        ("WARNING", 30),
        ("ERROR", 40),
        ("CRITICAL", 50),
        ("FOO", 20),
    ],
)
def test_format_log_level_from_record(level: str | int, expected: str | int):
    """Test the format_log_level_from_record function"""
    record = logging.LogRecord(
        name="test",
        level=logging.DEBUG,
        pathname="test.py",
        lineno=1,
        msg="test",
        args=None,
        exc_info=None,
    )
    assert format_log_level_from_record(record) == "DEBUG"



"""
Standard logging configuration for which intercepts standard logger messages from imported
libraries to emit through the loguru handler.  Must be the first import on the entry point to
ensure the first execution of the standard library logger basicConfig method
"""

import logging
from sys import stderr

from loguru import logger

__all__ = ["logger"]

from pydantic import Field
from pydantic_settings import BaseSettings
from pydantic_settings import SettingsConfigDict


class LogConfiguration(BaseSettings):
    """"""

    log_level: str = Field(default="DEBUG", validation_alias="LOGURU_LEVEL")
    model_config = SettingsConfigDict(env_file=".env", env_file_encoding="utf-8", extra="ignore")

    log_level_name_int_map: dict[str, int] = Field(
        default={
            "TRACE": 5,
            "DEBUG": 10,
            "INFO": 20,
            "SUCCESS": 25,
            "WARNING": 30,
            "ERROR": 40,
            "CRITICAL": 50,
        }
    )

    @property
    def log_level_int(self):
        return self.log_level_name_int_map.get(self.log_level, 20)

    @property
    def is_better_exceptions_active(self):
        return self.log_level_int <= 10


logger_config = LogConfiguration()

# Remove the default stderr handler
logger.remove()

# Add back the stderr handler with the 'diagnose' flag set
logger.add(
    stderr, level=logger_config.log_level, diagnose=logger_config.is_better_exceptions_active
)


def format_log_level_from_record(record: logging.LogRecord) -> int | str:
    """Log name to support formatting if known, otherwise use the level number"""
    log_level_int_name_map = {v: k for k, v in logger_config.log_level_name_int_map.items()}
    return log_level_int_name_map.get(record.levelno, 20)


class InterceptHandler(logging.Handler):
    """
    Handler to route stdlib logs to loguru
    """

    def emit(self, record: logging.LogRecord):
        # Retrieve context where the logging call occurred, this happens to be in the 6th frame upward
        logger_opt = logger.opt(depth=6, exception=record.exc_info)  # pragma: no cover
        logger_opt.log(
            format_log_level_from_record(record), record.getMessage()
        )  # pragma: no cover


# Configuration for stdlib logger to route messages to loguru; must be run before other imports
logging.basicConfig(handlers=[InterceptHandler()], level=logger_config.log_level_int)
